# fmt: off
# Warning: Don't edit file (autogenerated from python -m dev codegen).
from typing import List

ROBOT_RUN_TEST = "robot.runTest"  # Run Test/Task
ROBOT_DEBUG_TEST = "robot.debugTest"  # Debug Test/Task
ROBOT_RUN_SUITE = "robot.runSuite"  # Run Tests/Tasks Suite
ROBOT_DEBUG_SUITE = "robot.debugSuite"  # Debug Tests/Tasks Suite
ROBOT_INTERACTIVE_SHELL = "robot.interactiveShell"  # Start Interactive Console
ROBOT_INTERNAL_RFINTERACTIVE_START = "robot.internal.rfinteractive.start"  # Create Interactive Console
ROBOT_INTERNAL_RFINTERACTIVE_EVALUATE = "robot.internal.rfinteractive.evaluate"  # Evaluate in Interactive Console
ROBOT_INTERNAL_RFINTERACTIVE_STOP = "robot.internal.rfinteractive.stop"  # Stop Interactive Console
ROBOT_INTERNAL_RFINTERACTIVE_SEMANTIC_TOKENS = "robot.internal.rfinteractive.semanticTokens"  # Get the semantic tokens based on the code entered.
ROBOT_INTERNAL_RFINTERACTIVE_RESOLVE_COMPLETION = "robot.internal.rfinteractive.resolveCompletion"  # Resolves the passed completion.
ROBOT_INTERNAL_RFINTERACTIVE_COMPLETIONS = "robot.internal.rfinteractive.completions"  # Get the completions based on the code entered.
ROBOT_GET_RFLS_HOME_DIR = "robot.getRFLSHomeDir"  # Provides the directory used to store information (usually ~/.robotframework-ls -- may be configured with `ROBOTFRAMEWORK_LS_USER_HOME` environment variable).
ROBOT_CLEAR_CACHES_AND_RESTART_PROCESSES = "robot.clearCachesAndRestartProcesses"  # Clear caches and restart Robot Framework Language Server
ROBOT_CLEAR_CACHES_AND_RESTART_PROCESSES_START_INTERNAL = "robot.clearCachesAndRestartProcesses.start.internal"  # Stops the RFLS and waits for robot.clearCachesAndRestartProcesses.finish.internal to restart
ROBOT_CLEAR_CACHES_AND_RESTART_PROCESSES_FINISH_INTERNAL = "robot.clearCachesAndRestartProcesses.finish.internal"  # To be used to restart the processes
ROBOT_START_INDEXING_INTERNAL = "robot.startIndexing.internal"  # Starts the indexing service
ROBOT_WAIT_FULL_TEST_COLLECTION_INTERNAL = "robot.waitFullTestCollection.internal"  # Schedules and Waits for a full test collection
ROBOT_RF_INFO_INTERNAL = "robot.rfInfo.internal"  # Collects information on the Robot Framework version being used.
ROBOT_LINT_WORKSPACE = "robot.lint.workspace"  # Lint all files in the workspace.
ROBOT_LINT_EXPLORER = "robot.lint.explorer"  # Lint
ROBOT_OPEN_FLOW_EXPLORER = "robot.openFlowExplorer"  # Open Robot Flow Explorer
ROBOT_OPEN_FLOW_EXPLORER_INTERNAL = "robot.openFlowExplorer.internal"  # Open Robot Flow Explorer Internals
ROBOT_GENERATE_FLOW_EXPLORER_MODEL = "robot.generateFlowExplorerModel"  # Generate Robot Flow Explorer Model
ROBOT_COLLECT_ROBOT_DOCUMENTATION = "robot.collectRobotDocumentation"  # Collect Robot Documentation
ROBOT_VIEW_DOCUMENTATION_PIN = "robot.view.documentation.pin"  # Pin Robot Documentation
ROBOT_VIEW_DOCUMENTATION_UNPIN = "robot.view.documentation.unpin"  # Unpin Robot Documentation
ROBOT_CONVERT_OUTPUT_XML_TO_ROBOSTREAM = "robot.convertOutputXMLToRobostream"  # Convert output.xml to robocorp stream
ROBOT_APPLY_CODE_ACTION = "robot.applyCodeAction"  # Applies some code action

ALL_SERVER_COMMANDS: List[str] = [
    ROBOT_INTERNAL_RFINTERACTIVE_START,
    ROBOT_INTERNAL_RFINTERACTIVE_EVALUATE,
    ROBOT_INTERNAL_RFINTERACTIVE_STOP,
    ROBOT_INTERNAL_RFINTERACTIVE_SEMANTIC_TOKENS,
    ROBOT_INTERNAL_RFINTERACTIVE_RESOLVE_COMPLETION,
    ROBOT_INTERNAL_RFINTERACTIVE_COMPLETIONS,
    ROBOT_GET_RFLS_HOME_DIR,
    ROBOT_START_INDEXING_INTERNAL,
    ROBOT_WAIT_FULL_TEST_COLLECTION_INTERNAL,
    ROBOT_RF_INFO_INTERNAL,
    ROBOT_LINT_WORKSPACE,
    ROBOT_LINT_EXPLORER,
    ROBOT_OPEN_FLOW_EXPLORER_INTERNAL,
    ROBOT_GENERATE_FLOW_EXPLORER_MODEL,
    ROBOT_COLLECT_ROBOT_DOCUMENTATION,
    ROBOT_CONVERT_OUTPUT_XML_TO_ROBOSTREAM,
    ROBOT_APPLY_CODE_ACTION,
]

# fmt: on
