# fmt: off
# Warning: Don't edit file (autogenerated from python -m dev codegen).

OPTION_ROBOT_LANGUAGE_SERVER_PYTHON = "robot.language-server.python"
OPTION_ROBOT_LANGUAGE_SERVER_ARGS = "robot.language-server.args"
OPTION_ROBOT_LANGUAGE_SERVER_TCP_PORT = "robot.language-server.tcp-port"
OPTION_ROBOT_PYTHON_EXECUTABLE = "robot.python.executable"
OPTION_ROBOT_PYTHON_ENV = "robot.python.env"
OPTION_ROBOT_VARIABLES = "robot.variables"
OPTION_ROBOT_LOAD_VARIABLES_FROM_ARGUMENTS_FILE = "robot.loadVariablesFromArgumentsFile"
OPTION_ROBOT_PYTHONPATH = "robot.pythonpath"
OPTION_ROBOT_LIBRARIES_LIBDOC_NEEDS_ARGS = "robot.libraries.libdoc.needsArgs"
OPTION_ROBOT_LIBRARIES_LIBDOC_PRE_GENERATE = "robot.libraries.libdoc.preGenerate"
OPTION_ROBOT_CODE_FORMATTER = "robot.codeFormatter"
OPTION_ROBOT_FLOW_EXPLORER_THEME = "robot.flowExplorerTheme"
OPTION_ROBOT_LINT_ROBOCOP_ENABLED = "robot.lint.robocop.enabled"
OPTION_ROBOT_LINT_ENABLED = "robot.lint.enabled"
OPTION_ROBOT_LINT_UNDEFINED_KEYWORDS = "robot.lint.undefinedKeywords"
OPTION_ROBOT_LINT_UNDEFINED_LIBRARIES = "robot.lint.undefinedLibraries"
OPTION_ROBOT_LINT_UNDEFINED_RESOURCES = "robot.lint.undefinedResources"
OPTION_ROBOT_LINT_UNDEFINED_VARIABLE_IMPORTS = "robot.lint.undefinedVariableImports"
OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS = "robot.lint.keywordCallArguments"
OPTION_ROBOT_LINT_KEYWORD_RESOLVES_TO_MULTIPLE_KEYWORDS = "robot.lint.keywordResolvesToMultipleKeywords"
OPTION_ROBOT_LINT_VARIABLES = "robot.lint.variables"
OPTION_ROBOT_LINT_IGNORE_VARIABLES = "robot.lint.ignoreVariables"
OPTION_ROBOT_LINT_IGNORE_ENVIRONMENT_VARIABLES = "robot.lint.ignoreEnvironmentVariables"
OPTION_ROBOT_LINT_UNUSED_KEYWORD = "robot.lint.unusedKeyword"
OPTION_ROBOT_COMPLETIONS_SECTION_HEADERS_FORM = "robot.completions.section_headers.form"
OPTION_ROBOT_COMPLETIONS_KEYWORDS_NOT_IMPORTED_ENABLE = "robot.completions.keywordsNotImported.enable"
OPTION_ROBOT_COMPLETIONS_KEYWORDS_NOT_IMPORTED_ADD_IMPORT = "robot.completions.keywordsNotImported.addImport"
OPTION_ROBOT_COMPLETIONS_KEYWORDS_FORMAT = "robot.completions.keywords.format"
OPTION_ROBOT_COMPLETIONS_KEYWORDS_PREFIX_IMPORT_NAME = "robot.completions.keywords.prefixImportName"
OPTION_ROBOT_COMPLETIONS_KEYWORDS_PREFIX_IMPORT_NAME_IGNORE = "robot.completions.keywords.prefixImportNameIgnore"
OPTION_ROBOT_COMPLETIONS_KEYWORDS_ARGUMENTS_SEPARATOR = "robot.completions.keywords.argumentsSeparator"
OPTION_ROBOT_WORKSPACE_SYMBOLS_ONLY_FOR_OPEN_DOCS = "robot.workspaceSymbolsOnlyForOpenDocs"
OPTION_ROBOT_EDITOR_4SPACES_TAB = "robot.editor.4spacesTab"
OPTION_ROBOT_QUICK_FIX_KEYWORD_TEMPLATE = "robot.quickFix.keywordTemplate"
OPTION_ROBOT_CODE_LENS_ENABLE = "robot.codeLens.enable"
OPTION_ROBOT_CODE_LENS_RUN_ENABLE = "robot.codeLens.run.enable"
OPTION_ROBOT_CODE_LENS_INTERACTIVE_CONSOLE_ENABLE = "robot.codeLens.interactiveConsole.enable"
OPTION_ROBOT_INTERACTIVE_CONSOLE_ARGUMENTS = "robot.interactiveConsole.arguments"
OPTION_ROBOT_LANGUAGE = "robot.language"
OPTION_ROBOT_RUN_PEEK_ERROR_LEVEL = "robot.run.peekError.level"
OPTION_ROBOT_RUN_PEEK_ERROR_SHOW_SUMMARY = "robot.run.peekError.showSummary"
OPTION_ROBOT_RUN_PEEK_ERROR_SHOW_ERRORS_IN_CALLERS = "robot.run.peekError.showErrorsInCallers"

ALL_ROBOT_OPTIONS = frozenset(
    (
        OPTION_ROBOT_LANGUAGE_SERVER_PYTHON,
        OPTION_ROBOT_LANGUAGE_SERVER_ARGS,
        OPTION_ROBOT_LANGUAGE_SERVER_TCP_PORT,
        OPTION_ROBOT_PYTHON_EXECUTABLE,
        OPTION_ROBOT_PYTHON_ENV,
        OPTION_ROBOT_VARIABLES,
        OPTION_ROBOT_LOAD_VARIABLES_FROM_ARGUMENTS_FILE,
        OPTION_ROBOT_PYTHONPATH,
        OPTION_ROBOT_LIBRARIES_LIBDOC_NEEDS_ARGS,
        OPTION_ROBOT_LIBRARIES_LIBDOC_PRE_GENERATE,
        OPTION_ROBOT_CODE_FORMATTER,
        OPTION_ROBOT_FLOW_EXPLORER_THEME,
        OPTION_ROBOT_LINT_ROBOCOP_ENABLED,
        OPTION_ROBOT_LINT_ENABLED,
        OPTION_ROBOT_LINT_UNDEFINED_KEYWORDS,
        OPTION_ROBOT_LINT_UNDEFINED_LIBRARIES,
        OPTION_ROBOT_LINT_UNDEFINED_RESOURCES,
        OPTION_ROBOT_LINT_UNDEFINED_VARIABLE_IMPORTS,
        OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS,
        OPTION_ROBOT_LINT_KEYWORD_RESOLVES_TO_MULTIPLE_KEYWORDS,
        OPTION_ROBOT_LINT_VARIABLES,
        OPTION_ROBOT_LINT_IGNORE_VARIABLES,
        OPTION_ROBOT_LINT_IGNORE_ENVIRONMENT_VARIABLES,
        OPTION_ROBOT_LINT_UNUSED_KEYWORD,
        OPTION_ROBOT_COMPLETIONS_SECTION_HEADERS_FORM,
        OPTION_ROBOT_COMPLETIONS_KEYWORDS_NOT_IMPORTED_ENABLE,
        OPTION_ROBOT_COMPLETIONS_KEYWORDS_NOT_IMPORTED_ADD_IMPORT,
        OPTION_ROBOT_COMPLETIONS_KEYWORDS_FORMAT,
        OPTION_ROBOT_COMPLETIONS_KEYWORDS_PREFIX_IMPORT_NAME,
        OPTION_ROBOT_COMPLETIONS_KEYWORDS_PREFIX_IMPORT_NAME_IGNORE,
        OPTION_ROBOT_COMPLETIONS_KEYWORDS_ARGUMENTS_SEPARATOR,
        OPTION_ROBOT_WORKSPACE_SYMBOLS_ONLY_FOR_OPEN_DOCS,
        OPTION_ROBOT_EDITOR_4SPACES_TAB,
        OPTION_ROBOT_QUICK_FIX_KEYWORD_TEMPLATE,
        OPTION_ROBOT_CODE_LENS_ENABLE,
        OPTION_ROBOT_CODE_LENS_RUN_ENABLE,
        OPTION_ROBOT_CODE_LENS_INTERACTIVE_CONSOLE_ENABLE,
        OPTION_ROBOT_INTERACTIVE_CONSOLE_ARGUMENTS,
        OPTION_ROBOT_LANGUAGE,
        OPTION_ROBOT_RUN_PEEK_ERROR_LEVEL,
        OPTION_ROBOT_RUN_PEEK_ERROR_SHOW_SUMMARY,
        OPTION_ROBOT_RUN_PEEK_ERROR_SHOW_ERRORS_IN_CALLERS,
    )
)

# fmt: on
